"""
searchtree -- balanced search tree


Implementation based on [1]; especially, insert, delete, skew and split
methods of tree are just simple translations to Python.  Due to lack
of pointer in Python, we separate Node from Tree.

[1] Arne Anderson, 'Balanced Search Trees Made Simple.'
    In Proc. Workshop on Algorithms and Data Structures, pp. 60-71.
    Springer Verlag, 1993.
"""


class Node(object):
    """
    nodes of tree, searched by their keys.
    """
    ## # class variable
    ## bottom = None

    def __init__(self, key, value, level):
        """
        Node(key, value, level)
        """
        self.key = key
        self.value = value
        self.level = level

    @classmethod
    def get_bottom(cls):
        """
        get bottom.
        # singleton is better but it seems impossible
        #get a singleton bottom
        """
        ## if cls.bottom is None:
        ##      cls.bottom = cls(None, None, 0)
        ## return cls.bottom
        return cls(None, None, 0)

    def is_bottom(self):
        """
        Return True if the object is identical to bottom
        """
        return self.level == 0

    def __nonzero__(self):
        """
        Only bottom is False
        """
        return not self.is_bottom()

    def copy_except_level(self, original):
        """
        Copy the contents of original node into self except level.
        """
        self.key = original.key
        self.value = original.value

    # comparisons
    def __lt__(self, other):
        """
        inequality for sort
        """
        if hasattr(other, "key"):
            return self.key < other.key
        else:
            return self.key < other

    def __le__(self, other):
        """
        inequality for sort
        """
        if hasattr(other, "key"):
            return self.key <= other.key
        else:
            return self.key <= other

    def __eq__(self, other):
        """
        inequality for sort
        """
        if hasattr(other, "key"):
            return self.key == other.key
        else:
            return self.key == other

    def __ne__(self, other):
        """
        inequality for sort
        """
        if hasattr(other, "key"):
            return self.key != other.key
        else:
            return self.key != other

    def __ge__(self, other):
        """
        inequality for sort
        """
        if hasattr(other, "key"):
            return self.key >= other.key
        else:
            return self.key >= other

    def __gt__(self, other):
        """
        inequality for sort
        """
        if hasattr(other, "key"):
            return self.key > other.key
        else:
            return self.key > other


class Tree(object):
    """
    node of tree
    """
    def __init__(self, node, left, right):
        """
        """
        self.node = node
        self.left = left
        self.right = right
        self.node_class = node.__class__

    @classmethod
    def get_empty(cls, node_cls=Node):
        """
        Get an empty tree
        """
        bottom_tree = cls(node_cls.get_bottom(), None, None)
        bottom_tree.left = bottom_tree.right = bottom_tree
        return bottom_tree

    @classmethod
    def init_level_one(cls, key, value, node_cls=Node):
        """
        Initialize level one tree
        """
        node = node_cls(key, value, 1)
        return cls(node, cls.get_empty(node_cls), cls.get_empty(node_cls))

    @classmethod
    def new_tree(cls, top, node_cls=Node):
        """
        Create a new tree from a subtree rooted by top
        """
        tree = cls.get_empty(node_cls)
        tree.copy_from(top)
        return tree

    def is_bottom(self):
        """
        Return True if the object is identical to bottom
        """
        return self.node.is_bottom()

    def __nonzero__(self):
        """
        Only bottom is False
        """
        return not self.is_bottom()

    def is_leaf(self):
        """
        Return True if the object has bottom in both right and left.
        """
        return self.right.is_bottom() and self.left.is_bottom()

    def copy_from(self, original):
        """
        Copy all contents of the original.
        (to emulate pointer)
        """
        self.node = original.node
        self.left = original.left
        self.right = original.right
        self.node_class = original.node_class

    def _get_level(self):
        """
        return node.level
        """
        return self.node.level

    def _set_level(self, level):
        """
        modify node.level
        """
        self.node.level = level

    level = property(_get_level, _set_level)

    def skew(self):
        """
        skew procedure
        """
        if self.left.level == self.level:
            # rotate right
            temp = self.new_tree(self, self.node_class)
            self.copy_from(self.left)
            temp.left = self.right
            self.right = temp

    def split(self):
        """
        skew procedure

        Return a new tree.
        """
        if self.right.right.level == self.level:
            # rotate left
            temp = self.new_tree(self, self.node_class)
            self.copy_from(self.right)
            temp.right = self.left
            self.left = temp
            self.level += 1

    def insert(self, key, value):
        """
        insert a new data.
        """
        if self.is_bottom():
            self.node = self.node_class(key, value, 1)
            self.left = self.get_empty(self.node_class)
            self.right = self.get_empty(self.node_class)
        else:
            if self.node > key:
                self.left.insert(key, value)
            elif self.node < key:
                self.right.insert(key, value)
            self.skew()
            self.split()

    def delete(self, key, last=None, deleted=None):
        """
        delete a node of data
        """
        if not self.is_bottom():
            # search down the tree and set pointers last and deleted
            if self.node > key:
                last, deleted = self.left.delete(key, self, deleted)
            else:
                last, deleted = self.right.delete(key, self, self)

            # at the bottom of the tree we remove the element
            if (self == last and
                deleted and deleted.node == key):
                deleted.node.copy_except_level(self.node)
                deleted = self.get_empty(self.node_class)
                self.copy_from(self.right)
                last = None # dispose

            # on the way back, we balance
            elif (self.left.level < self.level - 1 or
                  self.right.level < self.level - 1):
                self.level -= 1
                if self.right.level >= self.level:
                    self.right.level = self.level
                self.skew()
                self.right.skew()
                self.right.right.skew()
                self.split()
                self.right.split()

        return last, deleted

    def search(self, key):
        """
        Return a node whose key is equal to the given key.

        If the key is not found in the tree, the result is a bottom
        node.
        """
        if self.is_bottom() or self.node == key:
            return self
        elif self.node < key:
            return self.right.search(key)
        else:
            return self.left.search(key)

    def search_interval(self, key, tentative=None):
        """
        Return a tuple of nodes (pred, hit, succ).

        pred is the node with the maximum key of lower bounds,
        and succ is the node with the minimum key of upper bounds.
        hit is, of course, the node with key.
        However, sometimes searching for key fails, then hit is
        a bottom node.  pred (or succ) also may be a bottom node,
        when key is less than or equal to the minimum node in tree
        (or more than or equal to the maximum, respectively).
        """
        if tentative is None:
            tentative = [self.get_empty(self.node_class),
                         None,
                         self.get_empty(self.node_class)]

        if self.is_bottom():
            tentative[1] = self.get_empty(self.node_class)
            return tuple(tentative)
        elif self.node < key:
            tentative[0] = self
            return self.right.search_interval(key, tentative)
        elif self.node > key:
            tentative[2] = self
            return self.left.search_interval(key, tentative)
        else: # if self.node == key:
            tentative[1] = self
            if self.left:
                tentative[0] = self.left.max()
            if self.right:
                tentative[2] = self.right.min()
            return tuple(tentative)

    def min(self):
        """
        Return a node with minimum key.
        """
        if self.left:
            return self.left.min()
        else:
            return self

    def max(self):
        """
        Return a node with maximum key.
        """
        if self.right:
            return self.right.max()
        else:
            return self

    def __len__(self):
        """
        Return the number of non-bottom nodes.

        This is just for test (very slow): O(n logn)
        """
        _len = 0
        left = self.min()
        while left:
            _len += 1
            _, hit, succ = self.search_interval(left.key + 1)
            if hit:
                left = hit
            elif succ:
                left = succ
            else:
                break
        return _len


class LinkedTree(Tree):
    """
    binary search tree which also linked as list.
    """
    def __init__(self, node, left, right, pred, succ):
        """
        init as Tree and add pred and succ
        """
        Tree.__init__(self, node, left, right)
        self.pred = pred
        self.succ = succ
        if pred is not None:
            pred.succ = self
        if succ is not None:
            succ.pred = self

    @classmethod
    def get_empty(cls, node_cls=Node):
        """
        Get an empty tree
        """
        bottom_tree = cls(node_cls.get_bottom(), None, None, None, None)
        bottom_tree.left = bottom_tree.right = bottom_tree
        return bottom_tree

    @classmethod
    def new_tree(cls, top, node_cls=Node):
        """
        Create a new tree from a subtree rooted by top
        """
        tree = cls.get_empty(node_cls)
        tree.copy_from(top)
        if top.pred is not None:
            top.pred.succ = tree
        if top.succ is not None:
            top.succ.pred = tree
        return tree

    @staticmethod
    def _link_nodes(pred, succ):
        """
        link two nodes pred and succ; current links are forgotten.
        """
        if pred:
            pred.succ = succ
        if succ:
            succ.pred = pred
        assert not pred or not succ or pred.node < succ.node

    def __len__(self):
        """
        Return the number of non-bottom nodes.

        This is just for test: O(n)
        """
        _len = 0
        left = self.min()
        while left:
            _len += 1
            left = left.succ
        return _len

    def copy_from(self, original):
        """
        Copy all contents of the original.
        (to emulate pointer)
        """
        Tree.copy_from(self, original)
        if original.pred != self:
            self.pred = original.pred
        if self.pred is not None:
            self.pred.succ = self
        if original.succ != self:
            self.succ = original.succ
        if self.succ is not None:
            self.succ.pred = self
        assert self.succ != self
        assert self.pred != self

    def insert_and_link(self, key, value, pred=None, succ=None):
        """
        insert a new data in the tree and list.
        """
        if self.is_bottom():
            self.node = self.node_class(key, value, 1)
            self.left = self.get_empty(self.node_class)
            self.right = self.get_empty(self.node_class)
            self._link_nodes(pred, self)
            self._link_nodes(self, succ)
        else:
            if self.node < key:
                self.right.insert_and_link(key, value, self, succ)
            elif self.node > key:
                self.left.insert_and_link(key, value, pred, self)
            else: # if self.node == key:
                raise ValueError("key %d is already in tree" % self.node.key)
            self.skew()
            self.split()

    def delete_and_unlink(self, key, last=None, deleted=None):
        """
        delete a subtree of data
        """
        if not self.is_bottom():
            # search down the tree and set pointers last and deleted
            if self.node > key:
                last, deleted = self.left.delete_and_unlink(key, self, deleted)
            else:
                last, deleted = self.right.delete_and_unlink(key, self, self)

            # at the bottom of the tree we remove the element
            if (self == last and
                deleted and deleted.node == key):
                # link keeping operations
                if deleted != self:
                    # copy last(=self) over deleted
                    deleted.node.copy_except_level(self.node) # rewrite!
                    self._link_nodes(deleted, self.succ)
                else:
                    self._link_nodes(deleted.pred, deleted.succ)
                self.pred = self.succ = None
                deleted = None # forget
                self.copy_from(self.right)
                last = None # forget

            # on the way back, we balance
            elif (self.left.level < self.level - 1 or
                  self.right.level < self.level - 1):
                self.level -= 1
                if self.right.level >= self.level:
                    self.right.level = self.level
                self.skew()
                self.right.skew()
                self.right.right.skew()
                self.split()
                self.right.split()

        return last, deleted

    def search_interval(self, key, tentative=None):
        """
        Return a tuple of subtrees (pred, hit, succ).

        pred is the tree with the maximum key of lower bounds,
        and succ is the tree with the minimum key of upper bounds.
        hit is, of course, the tree with the given key.
        However, sometimes searching for key fails, then hit is
        an empty tree.  pred (or succ) also may be an empty tree,
        when key is less than or equal to the minimum in tree
        (or more than or equal to the maximum, respectively).
        """
        if tentative is None:
            tentative = [self.get_empty(self.node_class),
                         None,
                         self.get_empty(self.node_class)]

        if self.is_bottom():
            tentative[1] = self.get_empty(self.node_class)
            return tuple(tentative)
        elif self.node < key:
            tentative[0] = self
            return self.right.search_interval(key, tentative)
        elif self.node > key:
            tentative[2] = self
            return self.left.search_interval(key, tentative)
        else: # if self.node == key:
            tentative[1] = self
            if self.pred:
                tentative[0] = self.pred
            elif self.left: # suspective
                tentative[0] = self.left.max()
            if self.succ:
                tentative[2] = self.succ
            elif self.right: # suspective
                tentative[2] = self.right.min()
            return tuple(tentative)
