"""
segment --- minimum cost calculation with zero cost segments
"""

import bisect
import collections
import heapq


Slope = collections.namedtuple('Slope', ['y', 'x', 'length', 'initval'])
Segment = collections.namedtuple('Segment', ['x', 'y', 'length'])


def minimum(string_length, segments):
    """
    Return minimum cost
    """
    if not segments:
        return string_length
    elif len(segments) == 1:
        return string_length - segments[0].length

    last = string_length, string_length
    return string_length - ValueGraph(string_length, segments)(*last)


class ValueGraph(object):
    """
    Value graph (is it an appropriate name?)
    """
    def __init__(self, string_length, slopes=None):
        self.string_length = string_length
        self.slopes = []
        if slopes is not None:
            for slope in slopes:
                self.add_slope(slope)

    def __call__(self, x, y):
        """
        Return the value at (x, y)
        """
        if not (0 <= x <= self.string_length and 0 <= y <= self.string_length):
            raise IndexError("graph index out of range: %d, %d" % (x, y))
        value = 0
        for slope in self.slopes:
            if x <= slope.x:
                continue
            if y <= slope.y:
                break
            if (slope.y < y <= slope.y + slope.length or
                slope.x < x <= slope.x + slope.length):
                value = max(value, slope.initval + min(x - slope.x, y - slope.y))
            else:
                value = max(value, slope.initval + slope.length)
        return value

    def add_slope(self, slope):
        """
        Add a slope to the graph.  Slopes are sorted by y coordinate.

        Given 'slope' can be either Slope or Segment.
        """
        new_slope = Slope(x=slope.x,
                          y=slope.y,
                          length=slope.length,
                          initval=self(slope.x, slope.y))
        bisect.insort_right(self.slopes, new_slope)


class ValueField(object):
    """
    Value field
    """
    def __init__(self, string_length, segments=None):
        self.string_length = string_length
        self.slopes = []
        if segments is not None:
            self.heap = list(segments)
            heapq.heapify(self.heap)
            while self.heap:
                segment = heapq.heappop(self.heap)
                maximum, support = self.maximum_support(segment)
                delay = self.start_delay(segment, support)
                if not delay:
                    self.add_slope(segment, maximum)
                else:
                    self.postpone(segment, delay)

    def __call__(self, x, y):
        """
        Get value at x, y
        """
        if not (0 <= x <= self.string_length and 0 <= y <= self.string_length):
            raise IndexError("graph index out of range: %d, %d" % (x, y))
        segment = Segment(x, y, 0)
        return self.maximum_support(segment)[0]

    def maximum_support(self, cand):
        """
        Return start value and slope that support the value.
        The support may be None when a slope provides the maximum,
        but does not have common x or y interval.
        """
        maximum, support = 0, None
        deletables = set()
        for slope in self.slopes:
            if cand.x <= slope.x:
                continue
            if cand.y <= slope.y:
                break
            if (slope.y < cand.y <= slope.y + slope.length or
                slope.x < cand.x <= slope.x + slope.length):
                value = slope.initval + min(cand.x - slope.x, cand.y - slope.y)
                if value > maximum:
                    maximum, support = value, slope
                elif (support and
                      support.initval + support.length > slope.initval + slope.length and
                      support.x + support.length >= slope.x + slope.length and
                      support.y + support.length <= slope.y + slope.length):
                    deletables.add(slope)
            else:
                value = slope.initval + slope.length
                if value > maximum:
                    maximum, support = value, None
                elif (support and
                      support.initval + support.length > slope.initval + slope.length and
                      support.x + support.length >= slope.x + slope.length and
                      support.y + support.length <= slope.y + slope.length):
                    deletables.add(slope)
        for slope in deletables:
            del self.slopes[bisect.bisect_left(self.slopes, slope)]
        return maximum, support

    @staticmethod
    def start_delay(cand, support):
        """
        Return start delay of cand.

        If support is None, it is zero.
        """
        diff = 0
        if support is not None:
            if support.y < cand.y <= support.y + support.length:
                diff = support.y + support.length - cand.y
            if support.x < cand.x <= support.x + support.length:
                diff_x = support.x + support.length - cand.x
                if diff_x > diff:
                    diff = diff_x
        return diff

    def add_slope(self, slope, initval):
        """
        Add a slope to the graph.  Slopes are sorted by y coordinate.

        Given 'slope' can be either Slope or Segment.
        """
        new_slope = Slope(x=slope.x,
                          y=slope.y,
                          length=slope.length,
                          initval=initval)
        bisect.insort_right(self.slopes, new_slope)

    def postpone(self, segment, delay):
        """
        Postpone to add segment to slopes for delay.
        """
        if delay < segment.length:
            delayed = Segment(segment.x + delay, segment.y + delay,
                              segment.length - delay)
            heapq.heappush(self.heap, delayed)
